@@ -2,15 +2,164 @@
  *
  * \brief Implementation of library-based communication
  */
+#include <errno.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <signal.h>
-#include <errno.h>
 #include "mdi.h"
 #include "mdi_lib.h"
 #include "mdi_global.h"
 #include "mdi_general.h"
+#include "mdi_plug_py.h"
+
+#ifdef _WIN32
+#include <windows.h>
+#else
+#include <dlfcn.h>
+#endif
+
+
+/*! \brief Launch an MDI plugin
+ *
+ */
+int library_launch_plugin(const char* plugin_name, const char* options, void* mpi_comm_ptr,
+                          MDI_Driver_node_callback_t driver_node_callback,
+                          void* driver_callback_object) {
+  int ret;
+  int driver_code_id = current_code;
+  code* this_code = get_code(driver_code_id);
+  MPI_Comm mpi_comm = *(MPI_Comm*) mpi_comm_ptr;
+
+  // Note: Eventually, should probably replace this code with libltdl
+  // Get the path to the plugin
+  char* plugin_path = malloc( PLUGIN_PATH_LENGTH * sizeof(char) );
+
+  // Get the name of the plugin's init function
+  char* plugin_init_name = malloc( PLUGIN_PATH_LENGTH * sizeof(char) );
+  snprintf(plugin_init_name, PLUGIN_PATH_LENGTH, "MDI_Plugin_init_%s", plugin_name);
+
+  // initialize a communicator for the driver
+  int icomm = library_initialize();
+  communicator* driver_comm = get_communicator(current_code, icomm);
+  library_data* libd = (library_data*) driver_comm->method_data;
+  libd->connected_code = (int)codes.size;
+
+  MDI_Comm comm;
+  ret = MDI_Accept_Communicator(&comm);
+  if ( ret != 0 || comm == MDI_COMM_NULL ) {
+    mdi_error("MDI unable to create communicator for plugin");
+    return -1;
+  }
+
+  // Set the driver callback function to be used by this plugin instance
+  libd->driver_callback_obj = driver_callback_object;
+  libd->driver_node_callback = driver_node_callback;
+
+  // Set the mpi communicator associated with this plugin instance
+  libd->mpi_comm = mpi_comm;
+
+
+  /*************************************************/
+  /*************** BEGIN PLUGIN MODE ***************/
+  /*************************************************/
+  plugin_mode = 1;
+
+  // Attempt to load a python script
+  snprintf(plugin_path, PLUGIN_PATH_LENGTH, "%s/%s.py", this_code->plugin_path, plugin_name);
+  if ( file_exists(plugin_path) ) {
+    ret = python_plugin_init( plugin_name, plugin_path, options, mpi_comm_ptr );
+    if ( ret != 0 ) {
+      mdi_error("Unable to initialize Python plugin");
+      return -1;
+    }
+  }
+  else {
+
+#ifdef _WIN32
+  // Attempt to open a library with a .dll extension
+  snprintf(plugin_path, PLUGIN_PATH_LENGTH, "%s/lib%s.dll", this_code->plugin_path, plugin_name);
+  HINSTANCE plugin_handle = LoadLibrary( plugin_path );
+  if ( ! plugin_handle ) {
+    // Unable to find the plugin library
+    free( plugin_path );
+    mdi_error("Unable to open MDI plugin");
+    return -1;
+  }
+
+  // Load a plugin's initialization function
+  MDI_Plugin_init_t plugin_init = (MDI_Plugin_init_t) (intptr_t) GetProcAddress( plugin_handle, plugin_init_name );
+  if ( ! plugin_init ) {
+    free( plugin_path );
+    free( plugin_init_name );
+    FreeLibrary( plugin_handle );
+    mdi_error("Unable to load MDI plugin init function");
+    return -1;
+  }
+
+#else
+  // Attempt to open a library with a .so extension
+  snprintf(plugin_path, PLUGIN_PATH_LENGTH, "%s/lib%s.so", this_code->plugin_path, plugin_name);
+  void* plugin_handle = dlopen(plugin_path, RTLD_NOW);
+  if ( ! plugin_handle ) {
+
+    // Attempt to open a library with a .dylib extension
+    snprintf(plugin_path, PLUGIN_PATH_LENGTH, "%s/lib%s.dylib", this_code->plugin_path, plugin_name);
+    plugin_handle = dlopen(plugin_path, RTLD_NOW);
+    if ( ! plugin_handle ) {
+      // Unable to find the plugin library
+      free( plugin_path );
+      free( plugin_init_name );
+      mdi_error("Unable to open MDI plugin");
+      return -1;
+    }
+  }
+
+  // Load a plugin's initialization function
+  MDI_Plugin_init_t plugin_init = (MDI_Plugin_init_t) (intptr_t) dlsym(plugin_handle, plugin_init_name);
+  if ( ! plugin_init ) {
+    free( plugin_path );
+    free( plugin_init_name );
+    dlclose( plugin_handle );
+    mdi_error("Unable to load MDI plugin init function");
+    return -1;
+  }
+#endif
+
+  // Initialize an instance of the plugin
+  ret = plugin_init();
+  if ( ret != 0 ) {
+    mdi_error("MDI plugin init function returned non-zero exit code");
+    return -1;
+  }
+
+  // Close the plugin library
+#ifdef _WIN32
+  FreeLibrary( plugin_handle );
+#else
+  dlclose( plugin_handle );
+#endif
+
+  }
+
+  /*************************************************/
+  /**************** END PLUGIN MODE ****************/
+  /*************************************************/
+  plugin_mode = 0;
+  current_code = driver_code_id;
+
+  // Delete the driver's communicator to the engine
+  // This will also delete the engine code and its communicator
+  delete_communicator(driver_code_id, comm);
+
+  // free memory from loading the plugin's initialization function
+  free( plugin_path );
+  free( plugin_init_name );
+
+  return 0;
+}
+
 
 /*! \brief Perform initialization of a communicator for library-based communication
  *
@@ -34,11 +183,33 @@ int library_initialize() {
   libd->connected_code = -1;
   libd->buf_allocated = 0;
   libd->execute_on_send = 0;
+  libd->mpi_comm = MPI_COMM_NULL;
   new_comm->method_data = libd;
 
+  // if this is an engine, go ahead and set the driver as the connected code
+  if ( strcmp(this_code->role, "ENGINE") == 0 ) {
+    int engine_code = current_code;
+    library_set_driver_current();
+    int driver_code_id = current_code;
+    libd->connected_code = driver_code_id;
+    current_code = engine_code;
+
+    // set the engine's mpi communicator
+    if ( plugin_mode ) {
+      code* driver_code = get_code(driver_code_id);
+      MDI_Comm matching_handle = library_get_matching_handle(comm_id);
+      communicator* driver_comm = get_communicator(driver_code->id, matching_handle);
+      library_data* driver_libd = (library_data*) driver_comm->method_data;
+      libd->mpi_comm = driver_libd->mpi_comm;
+      this_code->intra_MPI_comm = libd->mpi_comm;
+      MPI_Comm_rank( this_code->intra_MPI_comm, &this_code->intra_rank );
+    }
+  }
+
   return new_comm->id;
 }
 
+
 /*! \brief Set the driver as the current code
  *
  */
@@ -68,15 +239,18 @@ int library_set_driver_current() {
   return 0;
 }
 
+
 /*! \brief Perform LIBRARY method operations upon a call to MDI_Accept_Communicator
  *
  */
 int library_accept_communicator() {
-  // set the driver as the current code, if this is an ENGINE that is linked as a LIBRARY
-  library_set_driver_current();
-
-  // get the driver code
   code* this_code = get_code(current_code);
+  if ( this_code->called_set_execute_command_func && (! plugin_mode) ) {
+    // library codes are not permitted to call MDI_Accept_communicator after calling
+    // MDI_Set_execute_command_func, so assume that this call is being made by the driver
+    library_set_driver_current();
+  }
+  this_code = get_code(current_code);
 
   // if this is a DRIVER, check if there are any ENGINES that are linked to it
   if ( strcmp(this_code->role, "DRIVER") == 0 ) {
@@ -94,6 +268,7 @@ int library_accept_communicator() {
 	  //iengine = icode;
 	  iengine = other_code->id;
 	  found_engine = 1;
+	  break;
 	}
       }
     }
@@ -102,17 +277,8 @@ int library_accept_communicator() {
     if ( found_engine == 1 ) {
       int icomm = library_initialize();
 
-      // set the connected code for the engine
-      code* engine_code = get_code(iengine);
-      if ( engine_code->comms->size != 1 ) {
-	mdi_error("MDI_Accept_Communicator error: Engine appears to have been initialized multiple times"); 
-	return 1;
-      }
-      communicator* engine_comm = vector_get(engine_code->comms,0);
-      library_data* engine_lib = (library_data*) engine_comm->method_data;
-      engine_lib->connected_code = current_code;
-
       // set the connected code for the driver
+      code* engine_code = get_code(iengine);
       communicator* this_comm = get_communicator(current_code, icomm);
       library_data* libd = (library_data*) this_comm->method_data;
       libd->connected_code = engine_code->id;
@@ -255,6 +421,8 @@ int library_execute_command(MDI_Comm comm) {
  *                   2: The body (data) of a message.
  */
 int library_send(const void* buf, int count, MDI_Datatype datatype, MDI_Comm comm, int msg_flag) {
+  int ret;
+
   if ( datatype != MDI_INT && datatype != MDI_DOUBLE && datatype != MDI_CHAR && datatype != MDI_BYTE ) {
     mdi_error("MDI data type not recognized in library_send");
     return 1;
@@ -280,22 +448,9 @@ int library_send(const void* buf, int count, MDI_Datatype datatype, MDI_Comm com
 
     // determine the byte size of the data type being sent
     size_t datasize;
-    if (datatype == MDI_INT) {
-      datasize = sizeof(int);
-    }
-    else if (datatype == MDI_DOUBLE) {
-      datasize = sizeof(double);
-    }
-    else if (datatype == MDI_CHAR) {
-      datasize = sizeof(char);
-    }
-    else if (datatype == MDI_BYTE) {
-      datasize = sizeof(char);
-    }
-    else {
-      mdi_error("MDI Error: Unrecognized data type");
-      return 1;
-    }
+    MDI_Datatype basetype;
+    ret = datatype_info(datatype, &datasize, &basetype);
+    if ( ret != 0 ) { return ret; }
 
     int nheader_actual = 4; // actual number of elements of nheader that were sent
 
@@ -312,16 +467,16 @@ int library_send(const void* buf, int count, MDI_Datatype datatype, MDI_Comm com
       int body_type = header[2];
       int body_size = header[3];
       size_t body_stride;
-      if (datatype == MDI_INT) {
+      if (body_type == MDI_INT) {
 	body_stride = sizeof(int);
       }
-      else if (datatype == MDI_DOUBLE) {
+      else if (body_type == MDI_DOUBLE) {
 	body_stride = sizeof(double);
       }
-      else if (datatype == MDI_CHAR) {
+      else if (body_type == MDI_CHAR) {
 	body_stride = sizeof(char);
       }
-      else if (datatype == MDI_BYTE) {
+      else if (body_type == MDI_BYTE) {
 	body_stride = sizeof(char);
       }
       else {
@@ -357,15 +512,6 @@ int library_send(const void* buf, int count, MDI_Datatype datatype, MDI_Comm com
       // copy the body into libd->buf
       memcpy((char*)libd->buf + offset, buf, datasize * count);
 
-      // check whether the recipient code should now execute its command
-      if ( libd->execute_on_send == 1 ) {
-	// have the recipient code execute its command
-	library_execute_command(comm);
-
-	// turn off the execute_on_send flag
-	libd->execute_on_send = 0;
-      }
-
     }
     else {
 
@@ -376,6 +522,15 @@ int library_send(const void* buf, int count, MDI_Datatype datatype, MDI_Comm com
 
   }
 
+  // check whether the recipient code should now execute its command
+  if ( msg_flag == 2 && libd->execute_on_send ) {
+    // have the recipient code execute its command
+    library_execute_command(comm);
+
+    // turn off the execute_on_send flag
+    libd->execute_on_send = 0;
+  }
+
   return 0;
 }
 
@@ -398,6 +553,8 @@ int library_send(const void* buf, int count, MDI_Datatype datatype, MDI_Comm com
  *                   2: The body (data) of a message.
  */
 int library_recv(void* buf, int count, MDI_Datatype datatype, MDI_Comm comm, int msg_flag) {
+  int ret;
+
   code* this_code = get_code(current_code);
   communicator* this = get_communicator(current_code, comm);
   library_data* libd = (library_data*) this->method_data;
@@ -426,21 +583,9 @@ int library_recv(void* buf, int count, MDI_Datatype datatype, MDI_Comm comm, int
 
   // determine the byte size of the data type being sent
   size_t datasize;
-  if (datatype == MDI_INT) {
-    datasize = sizeof(int);
-  }
-  else if (datatype == MDI_DOUBLE) {
-    datasize = sizeof(double);
-  }
-  else if (datatype == MDI_CHAR) {
-    datasize = sizeof(char);
-  }
-  else if (datatype == MDI_BYTE) {
-    datasize = sizeof(char);
-  }
-  else {
-    mdi_error("MDI Error: Unrecognized data type");
-  }
+  MDI_Datatype basetype;
+  ret = datatype_info(datatype, &datasize, &basetype);
+  if ( ret != 0 ) { return ret; }
 
   // confirm that libd->buf is initialized
   if ( other_lib->buf_allocated != 1 ) {
